"""Module to work with Siren actions."""

import enum

from lila.core.base import Component
from lila.core.field import Field


@enum.unique
class Method(enum.Enum):
    """Enumerable with supported methods."""
    GET = "GET"
    PUT = "PUT"
    POST = "POST"
    DELETE = "DELETE"
    PATCH = "PATCH"

    def __str__(self):
        return self.value


class Action(Component):
    """Class to work with Siren actions."""

    def __init__(
            self,
            name,
            target,
            classes=(),
            method=Method.GET,
            title=None,
            fields=(),
            media_type=None,
        ):
        # pylint: disable=too-many-arguments
        super(Action, self).__init__(classes=classes, title=title)

        self._name = str(name)
        self._target = str(target)

        try:
            self._method = Method(str(method))
        except ValueError:
            raise ValueError("Method '{0}' is not supported".format(method))

        if any(not isinstance(field, Field) for field in fields):
            raise ValueError("Some of the fields are of incompatible type")

        if len(set(field.name for field in fields)) != len(fields):
            raise ValueError("Some of the fields have the same name")

        self._fields = tuple(fields)

        if media_type is not None:
            media_type = str(media_type)
        elif self._fields:
            media_type = "application/x-www-form-urlencoded"
        self._media_type = media_type

    @property
    def name(self):
        """Name of the action."""
        return self._name

    @property
    def target(self):
        """Request target of the action."""
        return self._target

    @property
    def method(self):
        """Method of the action."""
        return self._method

    @property
    def fields(self):
        """Fields to pass to the action."""
        return tuple(self._fields)

    @property
    def media_type(self):
        """Media type of action's payload."""
        return self._media_type



"""Module with default marshaler for an entity."""

import logging
import json

from lila.core.entity import Entity, EmbeddedRepresentation


class EntityMarshaler:
    """Class to marshal a single entity."""

    def __init__(self, entity, marshaler):
        self._entity = entity
        self._marshaler = marshaler

    def marshal(self):
        """Marshal the entity.

        :returns: dictionary with entity data.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal an entity")

        entity_data = {
            "class": self.marshal_classes(),
            "properties": self.marshal_properties(),
            "entities": self.marshal_entities(),
            "links": self.marshal_links(),
            "actions": self.marshal_actions(),
            "title": self.marshal_title(),
            }

        logger.info("Successfully marshaled an entity")
        return entity_data

    def marshal_classes(self):
        """Marshal entity's classes.

        :returns: list with string names of entity's classes.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        entity = self._entity
        try:
            classes = list(str(class_) for class_ in entity.classes)
        except AttributeError as error:
            logger.error("Failed to get entity's classes")
            raise ValueError("Failed to get entity's classes") from error
        except TypeError as error:
            logger.error("Failed to marshal an entity: failed to iterate over entity's classes")
            raise ValueError("Failed to iterate over entity's classes") from error

        return classes

    def marshal_properties(self):
        """Marshal entity's properties.

        :returns: JSON serializable object with entity's properties.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        entity = self._entity
        try:
            properties = json.loads(json.dumps(entity.properties))
        except AttributeError as error:
            logger.error("Failed to get entity's")
            raise ValueError("Failed to get entity's properties") from error
        except TypeError as error:
            logger.error("Failed to marshal entity's properties")
            raise ValueError("Failed to marshal entity's properties") from error

        return properties

    def marshal_entities(self):
        """Marshal entity's sub-entities.

        :returns: list with marshaled data of entity's sub-entities.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        entity = self._entity
        try:
            entity_sub_entities = list(entity.entities)
        except AttributeError as error:
            logger.error("Failed to get sub-entities of the entity")
            raise ValueError("Failed to get sub-entities of the entity") from error
        except TypeError as error:
            logger.error("Failed to iterate over sub-entities of the entity")
            raise ValueError("Failed to iterate over sub-entities of the entity") from error

        marshaler = self._marshaler
        marshal_sub_entity = lambda sub_entity: _marshal_sub_entity(sub_entity, marshaler)

        marshaled_sub_entities = []
        for sub_entity in entity_sub_entities:
            try:
                sub_entity_data = marshal_sub_entity(sub_entity)
            except Exception as error:
                logger.error("Failed to marshal sub-entities of the entity")
                raise ValueError("Failed to marshal sub-entities of the entity") from error

            marshaled_sub_entities.append(sub_entity_data)

        return marshaled_sub_entities

    def marshal_links(self):
        """Marshal entity's links.

        :returns: list with marshaled data of entity's links.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        entity = self._entity
        try:
            entity_links = list(entity.links)
        except AttributeError as error:
            logger.error("Failed to get entity's links")
            raise ValueError("Failed to get entity's links") from error
        except TypeError as error:
            logger.error("Failed to iterate over entity's links")
            raise ValueError("Failed to iterate over entity's links") from error

        marshal_link = self._marshaler.marshal_link

        marshaled_links = []
        for link in entity_links:
            try:
                link_data = marshal_link(link)
            except Exception as error:
                logger.error("Failed to marshal entity's links")
                raise ValueError("Failed to marshal entity's links") from error

            marshaled_links.append(link_data)

        return marshaled_links

    def marshal_actions(self):
        """Marshal entity's actions.

        :returns: list with marshaled data of entity's actions.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        entity = self._entity
        try:
            entity_actions = list(entity.actions)
        except AttributeError as error:
            logger.error("Failed to get entity's actions")
            raise ValueError("Failed to get entity's actions") from error
        except TypeError as error:
            logger.error("Failed to iterate over entity's actions")
            raise ValueError("Failed to iterate over entity's actions") from error

        marshal_action = self._marshaler.marshal_action
        marshaled_actions = []
        for action in entity_actions:
            try:
                action_data = marshal_action(action)
            except Exception as error:
                logger.error("Failed to marshal entity's actions")
                raise ValueError("Failed to marshal entity's actions") from error

            marshaled_actions.append(action_data)

        return marshaled_actions

    def marshal_title(self):
        """Marshal entity's title.

        :returns: string title of the entity or None.
        :raises: :class:ValueError.
        """
        entity = self._entity
        try:
            title = entity.title
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get entity's title")
            raise ValueError("Failed to get entity's title") from error

        if title is not None:
            title = str(title)

        return title


class EmbeddedRepresentationMarshaler:
    """Class to marshal a single embedded representation."""

    def __init__(self, embedded_representation, marshaler):
        self._embedded_representation = embedded_representation
        self._marshaler = marshaler

    def marshal(self):
        """Marshal the embedded representation.

        :returns: dictionary with entity data.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal an embeddded representation")

        representation_data = {
            "rel": self.marshal_relations(),
            "class": self.marshal_classes(),
            "properties": self.marshal_properties(),
            "entities": self.marshal_entities(),
            "links": self.marshal_links(),
            "actions": self.marshal_actions(),
            "title": self.marshal_title(),
            }

        logger.info("Successfully marshaled an embedded representation")
        return representation_data

    def marshal_relations(self):
        """Marshal relations of the embedded representation.

        :returns: list of string relations of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        embedded_representation = self._embedded_representation
        try:
            relations = list(str(relation) for relation in embedded_representation.relations)
        except AttributeError as error:
            logger.error("Failed to get relations of the embedded representation")
            raise ValueError("Failed to get relations of the embedded representation") from error
        except TypeError as error:
            logger.error("Failed to iterate over relations of the embedded representation")
            raise ValueError(
                "Failed to iterate over relations of the embedded representation",
                ) from error

        return relations

    def marshal_classes(self):
        """Marshal classes of the embedded representation.

        :returns: list with string names of classes of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        embedded_representation = self._embedded_representation
        try:
            classes = list(str(class_) for class_ in embedded_representation.classes)
        except AttributeError as error:
            logger.error("Failed to get classes of the embedded representation")
            raise ValueError("Failed to get classes of the embedded representation") from error
        except TypeError as error:
            logger.error("Failed to iterate over classes of the embedded representation")
            raise ValueError(
                "Failed to iterate over classes of the embedded representation",
                ) from error

        return classes

    def marshal_properties(self):
        """Marshal properties of the embedded representation.

        :returns: JSON serializable object with properties of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        embedded_representation = self._embedded_representation
        try:
            properties = json.loads(json.dumps(embedded_representation.properties))
        except AttributeError as error:
            logger.error("Failed to get properties of the embedded representation")
            raise ValueError("Failed to get properties of the embedded representation") from error
        except TypeError as error:
            logger.error("Failed to marshal properties of the embedded representation")
            raise ValueError(
                "Failed to marshal properties of the embedded representation",
                ) from error

        return properties

    def marshal_entities(self):
        """Marshal sub-entities of the embedded representation.

        :returns: list with marshaled data of sub-entities of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        embedded_representation = self._embedded_representation
        try:
            representation_sub_entities = list(embedded_representation.entities)
        except AttributeError as error:
            logger.error("Failed to get sub-entities of the embedded representation")
            raise ValueError("Failed to get sub-entities of the embedded representation") from error
        except TypeError as error:
            logger.error("Failed to iterate over sub-entities of the embedded representation")
            raise ValueError(
                "Failed to iterate over sub-entities of the embedded representation",
                ) from error

        marshaler = self._marshaler
        marshal_sub_entity = lambda sub_entity: _marshal_sub_entity(sub_entity, marshaler)

        marshaled_sub_entities = []
        for sub_entity in representation_sub_entities:
            try:
                sub_entity_data = marshal_sub_entity(sub_entity)
            except Exception as error:
                logger.error("Failed to marshal sub-entities of the embedded representation")
                raise ValueError(
                    "Failed to marshal sub-entities of the embedded representation",
                    ) from error

            marshaled_sub_entities.append(sub_entity_data)

        return marshaled_sub_entities

    def marshal_links(self):
        """Marshal links of the embedded representation.

        :returns: list with marshaled data of links of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        embedded_representation = self._embedded_representation
        try:
            representation_links = list(embedded_representation.links)
        except AttributeError as error:
            logger.error("Failed to get links of the embedded representation")
            raise ValueError("Failed to get links of the embedded representation") from error
        except TypeError as error:
            logger.error("Failed to iterate over links of the embedded representation")
            raise ValueError(
                "Failed to iterate over links of the embedded representation",
                ) from error

        marshal_link = self._marshaler.marshal_link

        marshaled_links = []
        for link in representation_links:
            try:
                link_data = marshal_link(link)
            except Exception as error:
                logger.error("Failed to marshal links of the embedded representation")
                raise ValueError(
                    "Failed to marshal links of the embedded representation",
                    ) from error

            marshaled_links.append(link_data)

        return marshaled_links

    def marshal_actions(self):
        """Marshal actions of the embedded representation.

        :returns: list with marshaled data of actions of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        embedded_representation = self._embedded_representation
        try:
            representation_actions = list(embedded_representation.actions)
        except AttributeError as error:
            logger.error("Failed to get actions of the embedded representation")
            raise ValueError("Failed to get actions of the embedded representation") from error
        except TypeError as error:
            logger.error("Failed to iterate over actions of the embedded representation")
            raise ValueError(
                "Failed to iterate over actions of the embedded representation",
                ) from error

        marshal_action = self._marshaler.marshal_action

        marshaled_actions = []
        for action in representation_actions:
            try:
                action_data = marshal_action(action)
            except Exception as error:
                logger.error("Failed to marshal actions of the embedded representation")
                raise ValueError(
                    "Failed to marshal actions of the embedded representation",
                    ) from error

            marshaled_actions.append(action_data)

        return marshaled_actions

    def marshal_title(self):
        """Marshal title of the embedded representation.

        :returns: string title of the embedded representation or None.
        :raises: :class:ValueError.
        """
        embedded_representation = self._embedded_representation
        try:
            title = embedded_representation.title
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get title of the embedded representation")
            raise ValueError("Failed to get title of the embedded representation") from error

        if title is not None:
            title = str(title)

        return title


class EntityParser:
    """Class to parse a single entity."""

    def __init__(self, data, parser):
        self._data = data
        self._parser = parser

    def parse(self):
        """Parse the entity.

        :returns: :class:`Entity <lila.core.entity.Entity>`.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse an entity")

        entity_classes = self.parse_classes()
        entity_properties = self.parse_properties()
        entity_entities = self.parse_entities()
        entity_links = self.parse_links()
        entity_actions = self.parse_actions()
        entity_title = self.parse_title()

        try:
            entity = Entity(
                classes=entity_classes,
                properties=entity_properties,
                entities=entity_entities,
                links=entity_links,
                actions=entity_actions,
                title=entity_title,
                )
        except Exception as error:
            logger.error("Failed to create an entity with provided data")
            raise ValueError("Failed to create an entity with provided data") from error
        else:
            logger.info("Successfully parsed an entity")

        return entity

    def parse_classes(self):
        """Parse entity's classes.

        :returns: list with string names of entity's classes.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            entity_classes = self._data["class"]
        except TypeError as error:
            logger.error("Failed to get classes from entity data")
            raise ValueError("Failed to get classes from entity data") from error
        except KeyError:
            entity_classes = ()

        try:
            entity_classes = tuple(str(class_) for class_ in entity_classes)
        except TypeError as error:
            logger.error("Failed to iterate over classes from entity data")
            raise ValueError("Failed to iterate over classes from entity data") from error

        return entity_classes

    def parse_properties(self):
        """Parse entity's properties.

        :returns: JSON object with entity's properties.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            entity_properties = self._data["properties"]
        except TypeError as error:
            logger.error("Failed to get properties from entity data")
            raise ValueError("Failed to get properties from entity data") from error
        except KeyError:
            entity_properties = {}

        try:
            entity_properties = json.loads(json.dumps(entity_properties))
        except TypeError as error:
            logger.error("Failed to parse entity's properties")
            raise ValueError("Failed to parse entity's properties") from error

        return entity_properties

    def parse_entities(self):
        """Parse entity's sub-entities.

        :returns: list with parsed entity's sub-entities.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            entity_sub_entities_data = self._data["entities"]
        except TypeError as error:
            logger.error("Failed to get sub-entities data from entity data")
            raise ValueError("Failed to get sub-entities data from entity data") from error
        except KeyError:
            entity_sub_entities_data = ()

        try:
            entity_sub_entities_data = list(entity_sub_entities_data)
        except TypeError as error:
            logger.error("Failed to iterate over sub-entities data from entity data")
            raise ValueError("Failed to iterate over sub-entities data from entity data") from error

        parser = self._parser
        parse_sub_entity = lambda data: _parse_sub_entity(data, parser)

        entity_sub_entities = []
        for data in entity_sub_entities_data:
            try:
                sub_entity = parse_sub_entity(data)
            except Exception as error:
                logger.error("Failed to parse entity's sub-entities")
                raise ValueError("Failed to parse entity's sub-entities") from error

            entity_sub_entities.append(sub_entity)

        return tuple(entity_sub_entities)

    def parse_links(self):
        """Parse entity's links.

        :returns: list with parsed entity's links.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            entity_links_data = self._data["links"]
        except TypeError as error:
            logger.error("Failed to get links data from entity data")
            raise ValueError("Failed to get links data from entity data") from error
        except KeyError:
            entity_links_data = ()

        try:
            entity_links_data = list(entity_links_data)
        except TypeError as error:
            logger.error("Failed to iterate over links data from entity data")
            raise ValueError("Failed to iterate over links data from entity data") from error

        parse_link = self._parser.parse_link

        entity_links = []
        for data in entity_links_data:
            try:
                link = parse_link(data)
            except Exception as error:
                logger.error("Failed to parse entity's links")
                raise ValueError("Failed to parse entity's links") from error

            entity_links.append(link)

        return tuple(entity_links)

    def parse_actions(self):
        """Parse entity's actions.

        :returns: list with parsed entity's actions.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            entity_actions_data = self._data["actions"]
        except TypeError as error:
            logger.error("Failed to get actions data from entity data")
            raise ValueError("Failed to get actions data from entity data") from error
        except KeyError:
            entity_actions_data = ()

        try:
            entity_actions_data = list(entity_actions_data)
        except TypeError as error:
            logger.error("Failed to iterate over actions data from entity data")
            raise ValueError("Failed to iterate over actions data from entity data") from error

        parse_action = self._parser.parse_action

        entity_actions = []
        for data in entity_actions_data:
            try:
                action = parse_action(data)
            except Exception as error:
                logger.error("Failed to parse entity's actions")
                raise ValueError("Failed to parse entity's actions") from error

            entity_actions.append(action)

        return tuple(entity_actions)

    def parse_title(self):
        """Parse entity's title.

        :returns: string title of the entity or None.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            entity_title = self._data["title"]
        except TypeError as error:
            logger.error("Failed to get title from entity data")
            raise ValueError("Failed to get title from entity data") from error
        except KeyError:
            entity_title = None

        if entity_title is not None:
            entity_title = str(entity_title)

        return entity_title


class EmbeddedRepresentationParser:
    """Class to parse a single embedded representation."""

    def __init__(self, data, parser):
        self._data = data
        self._parser = parser

    def parse(self):
        """Parse the embedded representation.

        :returns: :class:`EmbeddedRepresentation <lila.core.entity.EmbeddedRepresentation>`.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse an embedded representation")

        representation_relations = self.parse_relations()
        representation_classes = self.parse_classes()
        representation_properties = self.parse_properties()
        representation_entities = self.parse_entities()
        representation_links = self.parse_links()
        representation_actions = self.parse_actions()
        representation_title = self.parse_title()

        try:
            representation = EmbeddedRepresentation(
                relations=representation_relations,
                classes=representation_classes,
                properties=representation_properties,
                entities=representation_entities,
                links=representation_links,
                actions=representation_actions,
                title=representation_title,
                )
        except Exception as error:
            logger.error("Failed to create an embedded representation with provided data")
            raise ValueError(
                "Failed to create an embedded representation with provided data",
                ) from error
        else:
            logger.info("Successfully parsed an embedded representation")

        return representation

    def parse_relations(self):
        """Parse relations of the embedded representation.

        :returns: list of string relations of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            representation_relations = self._data["rel"]
        except TypeError as error:
            logger.error("Failed to get relations from data of the embedded representation")
            raise ValueError(
                "Failed to get relations from data of the embedded representation",
                ) from error
        except KeyError as error:
            logger.error("Data of the embedded representation do not have required 'rel' key")
            raise ValueError(
                "Data of the embedded representation do not have required 'rel' key",
                ) from error

        try:
            representation_relations = tuple(str(relation) for relation in representation_relations)
        except TypeError as error:
            logger.error(
                "Failed to iterate over relations from data of the embedded representation",
                )
            raise ValueError(
                "Failed to iterate over relations from data of the embedded representation",
                ) from error

        return representation_relations

    def parse_classes(self):
        """Parse classes of the embedded representation.

        :returns: list with string names of classes of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            representation_classes = self._data["class"]
        except TypeError as error:
            logger.error("Failed to get classes from data of the embedded representation")
            raise ValueError(
                "Failed to get classes from data of the embedded representation",
                ) from error
        except KeyError:
            representation_classes = ()

        try:
            representation_classes = tuple(str(class_) for class_ in representation_classes)
        except TypeError as error:
            logger.error("Failed to iterate over classes from data of the embedded representation")
            raise ValueError(
                "Failed to iterate over classes from data of the embedded representation",
                ) from error

        return representation_classes

    def parse_properties(self):
        """Parse properties of the embedded representation.

        :returns: JSON object with properties of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            representation_properties = self._data["properties"]
        except TypeError as error:
            logger.error("Failed to get properties from data of the embedded representation")
            raise ValueError(
                "Failed to get properties from data of the embedded representation",
                ) from error
        except KeyError:
            representation_properties = {}

        try:
            representation_properties = json.loads(json.dumps(representation_properties))
        except TypeError as error:
            logger.error("Failed to parse properties of the embedded representation")
            raise ValueError("Failed to parse properties of the embedded representation") from error

        return representation_properties

    def parse_entities(self):
        """Parse sub-entities of the embedded representation.

        :returns: list with parsed sub-entities of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            representation_sub_entities_data = self._data["entities"]
        except TypeError as error:
            logger.error("Failed to get sub-entities data from data of the embedded representation")
            raise ValueError(
                "Failed to get sub-entities data from data of the embedded representation",
                ) from error
        except KeyError:
            representation_sub_entities_data = ()

        try:
            representation_sub_entities_data = list(representation_sub_entities_data)
        except TypeError as error:
            logger.error(
                "Failed to iterate over sub-entities data from data of the embedded representation",
                )
            raise ValueError(
                "Failed to iterate over sub-entities data from data of the embedded representation",
                ) from error

        parser = self._parser
        parse_sub_entity = lambda data: _parse_sub_entity(data, parser)

        representation_sub_entities = []
        for data in representation_sub_entities_data:
            try:
                sub_entity = parse_sub_entity(data)
            except Exception as error:
                logger.error("Failed to parse sub-entities of the embedded representation")
                raise ValueError(
                    "Failed to parse sub-entities of the embedded representation",
                    ) from error

            representation_sub_entities.append(sub_entity)

        return tuple(representation_sub_entities)

    def parse_links(self):
        """Parse links of the embedded representation.

        :returns: list with parsed links of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            representation_links_data = self._data["links"]
        except TypeError as error:
            logger.error("Failed to get links data from data of the embedded representation")
            raise ValueError(
                "Failed to get links data from data of the embedded representation",
                ) from error
        except KeyError:
            representation_links_data = ()

        try:
            representation_links_data = list(representation_links_data)
        except TypeError as error:
            logger.error(
                "Failed to iterate over links data from data of the embedded representation",
                )
            raise ValueError(
                "Failed to iterate over links data from data of the embedded representation",
                ) from error

        parse_link = self._parser.parse_link

        representation_links = []
        for data in representation_links_data:
            try:
                link = parse_link(data)
            except Exception as error:
                logger.error("Failed to parse links of the embedded representation")
                raise ValueError("Failed to parse links of the embedded representation") from error

            representation_links.append(link)

        return tuple(representation_links)

    def parse_actions(self):
        """Parse actions of the embedded representation.

        :returns: list with parsed actions of the embedded representation.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            representation_actions_data = self._data["actions"]
        except TypeError as error:
            logger.error(
                "Failed to get actions data from data of the embedded representation",
                )
            raise ValueError(
                "Failed to get actions data from data of the embedded representation",
                ) from error
        except KeyError:
            representation_actions_data = ()

        try:
            representation_actions_data = list(representation_actions_data)
        except TypeError as error:
            logger.error(
                "Failed to iterate over actions data from data of the embedded representation",
                )
            raise ValueError(
                "Failed to iterate over actions data from data of the embedded representation",
                ) from error

        parse_action = self._parser.parse_action

        representation_actions = []
        for data in representation_actions_data:
            try:
                action = parse_action(data)
            except Exception as error:
                logger.error("Failed to parse actions of the embedded representation")
                raise ValueError(
                    "Failed to parse actions of the embedded representation",
                    ) from error

            representation_actions.append(action)

        return tuple(representation_actions)

    def parse_title(self):
        """Parse title of the embedded representation.

        :returns: string title of the embedded representation or None.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            representation_title = self._data["title"]
        except TypeError as error:
            logger.error("Failed to get title from data of the embedded representation")
            raise ValueError(
                "Failed to get title from data of the embedded representation",
                ) from error
        except KeyError:
            representation_title = None

        if representation_title is not None:
            representation_title = str(representation_title)

        return representation_title


def _marshal_sub_entity(sub_entity, marshaler):
    """Marshal the sub-entity.

    :param sub_entity: either embedded link or embedded representation.
    :param marshaler: marshaller for the Siren entities.
    :returns: dictionary with sub-entity data.
    :raises: :class:ValueError.
    """
    logger = logging.getLogger(__name__)

    if hasattr(sub_entity, "target"):
        logger.debug("Marshal sub-entity as an embedded link")
        marshaled_sub_entity = marshaler.marshal_embedded_link(sub_entity)
    else:
        logger.debug("Marshal sub-entity as an embedded representation")
        marshaled_sub_entity = marshaler.marshal_embedded_representation(sub_entity)

    return marshaled_sub_entity


def _parse_sub_entity(data, parser):
    """Parse the sub-entity.

    :param data: dictionary with sub-entity data.
    :param parser: parser of the Siren entities.
    :returns: parsed sub-entity.
    :raises: :class:ValueError.
    """
    logger = logging.getLogger(__name__)

    if "href" in data:
        logger.debug("Parse data as for an embedded representation")
        parsed_sub_entity = parser.parse_embedded_link(data)
    else:
        logger.debug("Parse data as for an embedded link")
        parsed_sub_entity = parser.parse_embedded_representation(data)

    return parsed_sub_entity



"""Module with default marshaler and parser for a link and embedded link."""

import logging

from lila.core.link import Link, EmbeddedLink


class LinkMarshaler:
    """Class to marshal a single link."""

    def __init__(self, link):
        self._link = link

    def marshal(self):
        """Marshal the link.

        :returns: dictionary with link data.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal a link")

        link_data = {
            "rel": self.marshal_relations(),
            "class": self.marshal_classes(),
            "href": self.marshal_target(),
            "title": self.marshal_title(),
            "type": self.marshal_target_media_type(),
            }

        logger.info("Successfully marshaled a link")
        return link_data

    def marshal_relations(self):
        """Marshal link's relations.

        :returns: list of string relations of the link.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        link = self._link
        try:
            relations = list(str(relation) for relation in link.relations)
        except AttributeError as error:
            logger.error("Failed to get link's relations")
            raise ValueError("Failed to get link's relations") from error
        except TypeError as error:
            logger.error("Failed to iterate over link's relations")
            raise ValueError("Failed to iterate over link's relations") from error

        return relations

    def marshal_classes(self):
        """Marshal link's classes.

        :returns: list with string names of link's classes.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        link = self._link
        try:
            classes = list(str(class_) for class_ in link.classes)
        except AttributeError as error:
            logger.error("Failed to get link's classes")
            raise ValueError("Failed to get link's classes") from error
        except TypeError as error:
            logger.error("Failed to iterate over link's classes")
            raise ValueError("Failed to iterate over link's classes") from error

        return classes

    def marshal_target(self):
        """Marshal link's target.

        :returns: string target of the link.
        :raises: :class:ValueError.
        """
        link = self._link
        try:
            target = link.target
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get link's target")
            raise ValueError("Failed to get link's target") from error

        return str(target)

    def marshal_title(self):
        """Marshal link's title.

        :returns: string title of the link or None.
        :raises: :class:ValueError.
        """
        link = self._link
        try:
            title = link.title
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get link's title")
            raise ValueError("Failed to get link's title") from error

        if title is not None:
            title = str(title)

        return title

    def marshal_target_media_type(self):
        """Marshal link's target media type.

        :returns: string value of link's target media type or None.
        :raises: :class:ValueError.
        """
        link = self._link
        try:
            target_media_type = link.target_media_type
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get link's target media type")
            raise ValueError("Failed to get link's target media type") from error

        if target_media_type is not None:
            target_media_type = str(target_media_type)

        return target_media_type


class EmbeddedLinkMarshaler:
    """Class to marshal a single embedded link."""

    def __init__(self, embedded_link):
        self._embedded_link = embedded_link

    def marshal(self):
        """Marshal the embedded link.

        :returns: dictionary with data of the embedded link.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal an embedded link")

        embedded_link_data = {
            "rel": self.marshal_relations(),
            "class": self.marshal_classes(),
            "href": self.marshal_target(),
            "title": self.marshal_title(),
            "type": self.marshal_target_media_type(),
            }

        logger.info("Successfully marshaled an embedded link")
        return embedded_link_data

    def marshal_relations(self):
        """Marshal relations of the embedded link.

        :returns: list of string relations of the embedded link.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        embedded_link = self._embedded_link
        try:
            relations = list(str(relation) for relation in embedded_link.relations)
        except AttributeError as error:
            logger.error("Failed to get relations of the embedded link")
            raise ValueError("Failed to get relations of the embedded link") from error
        except TypeError as error:
            logger.error("Failed to iterate over relations of the embedded link")
            raise ValueError("Failed to iterate over relations of the embedded link") from error

        return relations

    def marshal_classes(self):
        """Marshal classes of the embedded link.

        :returns: list with string names of classes of the embedded link.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        embedded_link = self._embedded_link
        try:
            classes = list(str(class_) for class_ in embedded_link.classes)
        except AttributeError as error:
            logger.error("Failed to get classes of the embedded link")
            raise ValueError("Failed to get classes of the embedded link") from error
        except TypeError as error:
            logger.error("Failed to iterate over classes of the embedded link")
            raise ValueError("Failed to iterate over classes of the embedded link") from error

        return classes

    def marshal_target(self):
        """Marshal target of the embedded link.

        :returns: string target of the embedded link.
        :raises: :class:ValueError.
        """
        embedded_link = self._embedded_link
        try:
            target = embedded_link.target
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get target of the embedded link")
            raise ValueError("Failed to get target of the embedded link") from error

        return str(target)

    def marshal_title(self):
        """Marshal title of the embedded link.

        :returns: string title of the embedded link or None.
        :raises: :class:ValueError.
        """
        embedded_link = self._embedded_link
        try:
            title = embedded_link.title
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get title of the embedded link")
            raise ValueError("Failed to get title of the embedded link") from error

        if title is not None:
            title = str(title)

        return title

    def marshal_target_media_type(self):
        """Marshal target media type of the embedded link.

        :returns: string value of target media type of the embedded link or None.
        :raises: :class:ValueError.
        """
        embedded_link = self._embedded_link
        try:
            target_media_type = embedded_link.target_media_type
        except AttributeError as error:
            logging.getLogger(__name__).error(
                "Failed to get target media type of the embedded link",
                )
            raise ValueError("Failed to get target media type of the embedded link") from error

        if target_media_type is not None:
            target_media_type = str(target_media_type)

        return target_media_type


class LinkParser:
    """Class to parse a single link."""

    def __init__(self, data):
        self._data = data

    def parse(self):
        """Parse the link.

        :returns: :class:`Link <lila.core.link.Link>`.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse a link")

        link_relations = self.parse_relations()
        link_classes = self.parse_classes()
        link_target = self.parse_target()
        link_title = self.parse_title()
        link_target_media_type = self.parse_target_media_type()

        try:
            link = Link(
                relations=link_relations,
                classes=link_classes,
                target=link_target,
                title=link_title,
                target_media_type=link_target_media_type,
                )
        except Exception as error:
            logger.error("Failed to create a link with provided data")
            raise ValueError("Failed to create a link with provided data") from error
        else:
            logger.info("Successfully parsed a link")

        return link

    def parse_relations(self):
        """Parse link's relations.

        :returns: list of string relations of the link.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            link_relations = self._data["rel"]
        except TypeError as error:
            logger.error("Failed to get relations from link data")
            raise ValueError("Failed to get relations from link data") from error
        except KeyError as error:
            logger.error("Link data do not have required 'rel' key")
            raise ValueError("Link data do not have required 'rel' key") from error

        try:
            link_relations = tuple(str(relation) for relation in link_relations)
        except TypeError as error:
            logger.error("Failed to iterate over relations from link data")
            raise ValueError("Failed to iterate over relations from link data") from error

        return link_relations

    def parse_classes(self):
        """Parse link's classes.

        :returns: list with string names of link's classes.
        """
        logger = logging.getLogger(__name__)

        try:
            link_classes = self._data["class"]
        except TypeError as error:
            logger.error("Failed to get classes from link data")
            raise ValueError("Failed to get classes from link data") from error
        except KeyError:
            link_classes = ()

        try:
            link_classes = tuple(str(class_) for class_ in link_classes)
        except TypeError as error:
            logger.error("Failed to iterate over classes from link data")
            raise ValueError("Failed to iterate over classes from link data") from error

        return link_classes

    def parse_target(self):
        """Parse link's target.

        :returns: string target of the link.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            link_target = self._data["href"]
        except TypeError as error:
            logger.error("Failed to get target from link data")
            raise ValueError("Failed to get target from link data") from error
        except KeyError as error:
            logger.error("Link data do not have required 'href' key")
            raise ValueError("Link data do not have required 'href' key") from error

        return str(link_target)

    def parse_title(self):
        """Parse link's title.

        :returns: string title of the link or None.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            link_title = self._data["title"]
        except TypeError as error:
            logger.error("Failed to get title from link data")
            raise ValueError("Failed to get title from link data") from error
        except KeyError:
            link_title = None

        if link_title is not None:
            link_title = str(link_title)

        return link_title

    def parse_target_media_type(self):
        """Parse link's target media type.

        :returns: string value of link's target media type or None.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            link_target_media_type = self._data["type"]
        except TypeError as error:
            logger.error("Failed to get target media type from link data")
            raise ValueError("Failed to get target media type from link data") from error
        except KeyError:
            link_target_media_type = None

        if link_target_media_type is not None:
            link_target_media_type = str(link_target_media_type)

        return link_target_media_type


class EmbeddedLinkParser:
    """Class to parse a single embedded link."""

    def __init__(self, data):
        self._data = data

    def parse(self):
        """Parse the embedded link.

        :returns: :class:`EmbeddedLink <lila.core.link.Link>`.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse an embedded link")

        embedded_link_relations = self.parse_relations()
        embedded_link_classes = self.parse_classes()
        embedded_link_target = self.parse_target()
        embedded_link_title = self.parse_title()
        embedded_link_target_media_type = self.parse_target_media_type()

        try:
            embedded_link = EmbeddedLink(
                relations=embedded_link_relations,
                classes=embedded_link_classes,
                target=embedded_link_target,
                title=embedded_link_title,
                target_media_type=embedded_link_target_media_type,
                )
        except Exception as error:
            logger.error("Failed to create an embedded link with provided data")
            raise ValueError("Failed to create an embedded link with provided data") from error
        else:
            logger.info("Successfully parsed an embedded link")

        return embedded_link

    def parse_relations(self):
        """Parse relations of the embedded link.

        :returns: list of string relations of the embedded link.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            embedded_link_relations = self._data["rel"]
        except TypeError as error:
            logger.error("Failed to get relations from data of the embedded link")
            raise ValueError("Failed to get relations from data of the embedded link") from error
        except KeyError as error:
            logger.error("Data of the embedded link do not have required 'rel' key")
            raise ValueError("Data of the embedded link do not have required 'rel' key") from error

        try:
            embedded_link_relations = tuple(str(relation) for relation in embedded_link_relations)
        except TypeError as error:
            logger.error("Failed to iterate over relations from data of the embedded link")
            raise ValueError(
                "Failed to iterate over relations from data of the embedded link",
                ) from error

        return embedded_link_relations

    def parse_classes(self):
        """Parse classes of the embedded link.

        :returns: list with string names of classes of the embedded link.
        """
        logger = logging.getLogger(__name__)

        try:
            embedded_link_classes = self._data["class"]
        except TypeError as error:
            logger.error("Failed to get classes from data of the embedded link")
            raise ValueError("Failed to get classes from data of the embedded link") from error
        except KeyError:
            embedded_link_classes = ()

        try:
            embedded_link_classes = tuple(str(class_) for class_ in embedded_link_classes)
        except TypeError as error:
            logger.error("Failed to iterate over classes from data of the embedded link")
            raise ValueError(
                "Failed to iterate over classes from data of the embedded link",
                ) from error

        return embedded_link_classes

    def parse_target(self):
        """Parse target of the embedded link.

        :returns: string target of the embedded link.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            embedded_link_target = self._data["href"]
        except TypeError as error:
            logger.error("Failed to get target from data of the embedded link")
            raise ValueError("Failed to get target from data of the embedded link") from error
        except KeyError as error:
            logger.error("Data of the embedded link do not have required 'href' key")
            raise ValueError("Data of the embedded link do not have required 'href' key") from error

        return str(embedded_link_target)

    def parse_title(self):
        """Parse title of the embedded link.

        :returns: string title of the embedded link or None.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            embedded_link_title = self._data["title"]
        except TypeError as error:
            logger.error("Failed to get title from data of the embedded link")
            raise ValueError("Failed to get title from data of the embedded link") from error
        except KeyError:
            embedded_link_title = None

        if embedded_link_title is not None:
            embedded_link_title = str(embedded_link_title)

        return embedded_link_title

    def parse_target_media_type(self):
        """Parse target media type of the embedded link.

        :returns: string value of target media type of the embedded link or None.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            embedded_link_target_media_type = self._data["type"]
        except TypeError as error:
            logger.error("Failed to get target media type from data of the embedded link")
            raise ValueError(
                "Failed to get target media type from data of the embedded link",
                ) from error
        except KeyError:
            embedded_link_target_media_type = None

        if embedded_link_target_media_type is not None:
            embedded_link_target_media_type = str(embedded_link_target_media_type)

        return embedded_link_target_media_type



"""Module to work with Siren entities."""

from copy import deepcopy

import lila.core.common as common
from lila.core.base import Component
from lila.core.link import Link, EmbeddedLink
from lila.core.action import Action


class Entity(Component):
    """Class to work with Siren entities."""

    def __init__(self, title=None, classes=(), properties=(), entities=(), links=(), actions=()):
        # pylint: disable=too-many-arguments
        super(Entity, self).__init__(classes=classes, title=title)

        self._properties = common.adjust_properties(properties)

        if any(not isinstance(link, Link) for link in links):
            raise ValueError("Some of the links are of incompatible type")
        self._links = tuple(links)

        if any(not isinstance(action, Action) for action in actions):
            raise ValueError("Some of the actions are of incompatible type")

        if len(set(action.name for action in actions)) != len(actions):
            raise ValueError("Some of the actions have the same name")

        self._actions = tuple(actions)

        if any(not isinstance(entity, (EmbeddedLink, EmbeddedRepresentation)) for entity in entities):  # pylint: disable=line-too-long
            raise ValueError("Some of the entities are of incompatible type")
        self._entities = tuple(entities)

    @property
    def properties(self):
        """Properties of the entity."""
        return deepcopy(self._properties)

    @property
    def links(self):
        """Navigation links from the entity."""
        return tuple(self._links)

    @property
    def actions(self):
        """Actions of the entity."""
        return tuple(self._actions)

    @property
    def entities(self):
        """Subentities of the entity."""
        return tuple(self._entities)


class EmbeddedRepresentation(Entity):
    """Class to work with embedded Siren entities."""

    def __init__(
            self,
            relations,
            title=None,
            classes=(),
            properties=(),
            entities=(),
            links=(),
            actions=(),
        ):
        # pylint: disable=too-many-arguments
        relations = common.adjust_relations(relations)
        if not relations:
            raise ValueError("No relations are passed to create an embedded representation")
        self._relations = relations

        super(EmbeddedRepresentation, self).__init__(
            title=title,
            classes=classes,
            properties=properties,
            entities=entities,
            links=links,
            actions=actions,
            )

    @property
    def relations(self):
        """Relationship between the representation and parent entity."""
        return tuple(self._relations)



"""Module to work with Siren actions."""

import enum

from lila.core.base import Component


@enum.unique
class InputType(enum.Enum):
    """Enumerable with supported input types."""
    HIDDEN = "hidden"
    TEXT = "text"
    SEARCH = "search"
    PHONE = "tel"
    URL = "url"
    EMAIL = "email"
    PASSWORD = "password"
    DATETIME = "datetime"
    DATE = "date"
    MONTH = "month"
    WEEK = "week"
    TIME = "time"
    DATETIME_LOCAL = "datetime-local"
    NUMBER = "number"
    RANGE = "range"
    COLOR = "color"
    CHECKBOX = "ratio"
    FILE = "file"
    SUBMIT = "submit"
    IMAGE = "image"
    RESET = "reset"
    BUTTON = "button"

    def __str__(self):
        return self.value


class Field(Component):
    """Class to work with Siren fields."""

    def __init__(self, name, classes=(), input_type=InputType.TEXT, value=None, title=None):
        # pylint: disable=too-many-arguments
        super(Field, self).__init__(classes=classes, title=title)

        self._name = str(name)

        try:
            self._input_type = InputType(str(input_type))
        except ValueError:
            raise ValueError("Unsupported input type '{0}'".format(input_type))

        if value is not None:
            value = str(value)
        self._value = value

    @property
    def name(self):
        """Name of the field."""
        return self._name

    @property
    def input_type(self):
        """Input type of the field."""
        return self._input_type

    @property
    def value(self):
        """Value assigned to the field."""
        return self._value



"""Module to work with Siren links."""

import lila.core.common as common
from lila.core.base import Component


class Link(Component):
    """Class to work with Siren link."""

    def __init__(self, relations, target, classes=(), title=None, target_media_type=None):
        # pylint: disable=too-many-arguments
        super(Link, self).__init__(classes=classes, title=title)

        self._relations = common.adjust_relations(relations)
        self._target = str(target)

        if target_media_type is not None:
            target_media_type = str(target_media_type)
        self._target_media_type = target_media_type

    @property
    def relations(self):
        """Relationships between the link and entity."""
        return tuple(self._relations)

    @property
    def target(self):
        """Target of the link."""
        return self._target

    @property
    def target_media_type(self):
        """Media type of the target resource."""
        return self._target_media_type


class EmbeddedLink(Link):
    """Class to work with embedded Siren links."""

    def __init__(self, relations, target, classes=(), title=None, target_media_type=None):
        # pylint: disable=too-many-arguments
        relations = common.adjust_relations(relations)
        if not relations:
            raise ValueError("No relations are passed to create an embedded link")

        super(EmbeddedLink, self).__init__(
            relations=relations,
            target=target,
            classes=classes,
            title=title,
            target_media_type=target_media_type,
            )



"""Module with default marshaler and parser for a field."""

import logging

from lila.core.field import Field, InputType


class FieldMarshaler:
    """Class to marshal a single field."""

    def __init__(self, field):
        self._field = field

    def marshal(self):
        """Marshal the field.

        :returns: dictionary with field data.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal a field")

        field_data = {
            "name": self.marshal_name(),
            "class": self.marshal_classes(),
            "type": self.marshal_input_type(),
            "value": self.marshal_value(),
            "title": self.marshal_title(),
            }

        logger.info("Successfully marshaled a field")
        return field_data

    def marshal_name(self):
        """Marshal field's name.

        :returns: string name of the field.
        :raises: :class:ValueError.
        """
        field = self._field
        try:
            name = field.name
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get field's name")
            raise ValueError("Failed to get field's name") from error

        return str(name)

    def marshal_classes(self):
        """Marshal field's classes.

        :returns: list with string names of field's classes.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        field = self._field
        try:
            classes = list(str(class_) for class_ in field.classes)
        except AttributeError as error:
            logger.error("Failed to get field's classes")
            raise ValueError("Failed to get field's classes") from error
        except TypeError as error:
            logger.error("Failed to iterate over field's classes")
            raise ValueError("Failed to iterate over field's classes") from error

        return classes

    def marshal_input_type(self):
        """Marshal field's input type.

        :returns: string value of field's input type.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        field = self._field
        try:
            input_type = str(field.input_type)
        except AttributeError as error:
            logger.error("Failed to get field's input type")
            raise ValueError("Failed to get field's input type") from error

        try:
            input_type = InputType(input_type)
        except ValueError as error:
            logger.error("Field's input type is not supported")
            raise ValueError("Field's input type is not supported") from error

        return input_type.value

    def marshal_value(self):
        """Marshal field's value.

        :returns: string value of the field or None.
        :raises: :class:ValueError.
        """
        field = self._field
        try:
            value = field.value
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get field's value")
            raise ValueError("Failed to get field's value") from error

        if value is not None:
            value = str(value)

        return value

    def marshal_title(self):
        """Marshal field's title.

        :returns: string title of the field or None.
        :raises: :class:ValueError.
        """
        field = self._field
        try:
            title = field.title
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get field's title")
            raise ValueError("Failed to get field's title") from error

        if title is not None:
            title = str(title)

        return title


class FieldParser:
    """Class to parse a single field."""

    def __init__(self, data):
        self._data = data

    def parse(self):
        """Parse a field from the data.

        :returns: :class:`Field <lila.core.field.Field>`.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse a field")

        field_name = self.parse_name()
        field_classes = self.parse_classes()
        field_input_type = self.parse_input_type()
        field_value = self.parse_value()
        field_title = self.parse_title()

        try:
            field = Field(
                name=field_name,
                classes=field_classes,
                input_type=field_input_type,
                value=field_value,
                title=field_title,
                )
        except Exception as error:
            logger.error("Failed to create a field with provided data")
            raise ValueError("Failed to create a field with provided data") from error
        else:
            logger.info("Successfully parsed a field")

        return field

    def parse_name(self):
        """Parse field's name.

        :returns: string name of the field.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            field_name = self._data["name"]
        except TypeError as error:
            logger.error("Failed to get name from field data")
            raise ValueError("Failed to get name from field data") from error
        except KeyError as error:
            logger.error("Field data do not have required 'name' key")
            raise ValueError("Field data do not have required 'name' key") from error

        return str(field_name)

    def parse_classes(self):
        """Parse field's classes.

        :returns: list with string names of field's classes.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            field_classes = self._data["class"]
        except TypeError as error:
            logger.error("Failed to get classes from field data")
            raise ValueError("Failed to get classes from field data") from error
        except KeyError:
            field_classes = ()

        try:
            field_classes = tuple(str(class_) for class_ in field_classes)
        except TypeError as error:
            logger.error("Failed to iterate over classes from field data")
            raise ValueError("Failed to iterate over classes from field data") from error

        return field_classes

    def parse_input_type(self):
        """Parse field's input type.

        :returns: :class:`InputType <lila.core.field.InputType>`.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            field_input_type = self._data["type"]
        except TypeError as error:
            logger.error("Failed to get input type from field data")
            raise ValueError("Failed to get input type from field data") from error
        except KeyError:
            field_input_type = InputType.TEXT.value

        try:
            field_input_type = InputType(field_input_type)
        except ValueError as error:
            logger.error("Field data contain not supported input type")
            raise ValueError("Field data contain not supported input type") from error

        return field_input_type

    def parse_value(self):
        """Parse field's value.

        :returns: string value or None.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            field_value = self._data["value"]
        except TypeError as error:
            logger.error("Failed to get value from field data")
            raise ValueError("Failed to get value from field data") from error
        except KeyError:
            field_value = None

        if field_value is not None:
            field_value = str(field_value)

        return field_value

    def parse_title(self):
        """Parse field's title.

        :returns: string title of the field or None.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            field_title = self._data["title"]
        except TypeError as error:
            logger.error("Failed to get title from field data")
            raise ValueError("Failed to get title from field data") from error
        except KeyError:
            field_title = None

        if field_title is not None:
            field_title = str(field_title)

        return field_title



"""Module with common functions."""

import json


def adjust_classes(classes):
    """Adjust classes to Siren protocol.

    :param classes: iterable with classes.
    :returns: tuple with string names.
    :raises: :class:ValueError.
    """
    try:
        return tuple(str(class_) for class_ in classes)
    except TypeError as error:
        raise ValueError("Classes must be iterable with string values") from error


def adjust_relations(relations):
    """Adjust relations to Siren protocol.

    :param relations: iterable with relations.
    :returns: tuple with string relations.
    :raises: :class:ValueError.
    """
    try:
        return tuple(str(relation) for relation in relations)
    except TypeError as error:
        raise ValueError("Relations must be iterable with string values") from error


def adjust_properties(properties):
    """Adjust properties to Siren protocol.

    :param properties: dictionary or iterable with dictionary items.
    :returns: dictionary with strings as values (property names) and json serializable values.
    :raises: :class:ValueError.
    """
    try:
        properties_dictionary = dict(properties)
    except (TypeError, ValueError) as error:
        raise ValueError("Can't create dictionary from properties") from error

    adjusted_properties = {}
    for name, value in properties_dictionary.items():
        adjusted_name = str(name)
        # ensure that value is a valid JSON object by dumping and loading it.
        try:
            adjusted_value = json.loads(json.dumps(value))
        except TypeError as error:
            error_message = "Unsupported value for property '{name}'".format(name=adjusted_name)
            raise ValueError(error_message) from error

        adjusted_properties[adjusted_name] = adjusted_value

    return adjusted_properties



"""Module with JSON marshaler for Siren objects."""

import logging

from lila.serialization.marshaler import Marshaler
from lila.serialization.json.field import FieldMarshaler
from lila.serialization.json.action import ActionMarshaler
from lila.serialization.json.link import LinkMarshaler, EmbeddedLinkMarshaler
from lila.serialization.json.entity import EntityMarshaler, EmbeddedRepresentationMarshaler


class JSONMarshaler(Marshaler):
    """Class to marshal Siren objects into JSON."""

    create_field_marshaler = FieldMarshaler
    create_link_marshaler = LinkMarshaler
    create_embedded_link_marshaler = EmbeddedLinkMarshaler

    def create_action_marshaler(self, action):
        """Factory method to create a marshaler for an action.

        :param action: Siren action to marshal.
        :returns: :class:`ActionMarshaler <lila.serialization.json.action.ActionMarshaler>`.
        """
        return ActionMarshaler(action=action, marshaler=self)

    def create_entity_marshaler(self, entity):
        """Factory method to create a marshaler for an entity.

        :param entity: Siren entity to marshal.
        :returns: :class:`EntityMarshaler <lila.serialization.json.entity.EntityMarshaler>`.
        """
        return EntityMarshaler(entity=entity, marshaler=self)

    def create_embedded_representation_marshaler(self, embedded_representation):
        """Factory method to create a marshaler for an embedded representation.

        :param embedded_representation: Siren embedded representation to marshal.
        :returns: :class:`EmbeddedRepresentationMarshaler
            <lila.serialization.json.entity.EmbeddedRepresentationMarshaler>`.
        """
        return EmbeddedRepresentationMarshaler(
            embedded_representation=embedded_representation,
            marshaler=self,
            )

    def marshal_field(self, field):
        """Marshal Siren field.

        :param field: Siren Field.
        :returns: dictionary with field data.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal a field '%s'", field)

        marshaler = self.create_field_marshaler(field)
        try:
            marshaled_field = marshaler.marshal()
        except Exception:
            logger.error("Failed to marshal a field")
            raise

        logger.info("Successfully marshaled a field")
        return marshaled_field

    def marshal_action(self, action):
        """Marshal Siren action.

        :param action: Siren Action.
        :returns: dictionary with action data.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal an action '%s'", action)

        marshaler = self.create_action_marshaler(action)
        try:
            marshaled_action = marshaler.marshal()
        except Exception:
            logger.error("Failed to marshal an action")
            raise

        logger.info("Successfully marshaled an action")
        return marshaled_action

    def marshal_link(self, link):
        """Marshal Siren link.

        :param link: Siren Link.
        :returns: dictionary with link data.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal a link '%s'", link)

        marshaler = self.create_link_marshaler(link)
        try:
            marshaled_link = marshaler.marshal()
        except Exception:
            logger.error("Failed to marshal a link")
            raise

        logger.info("Successfully marshaled a link")
        return marshaled_link

    def marshal_embedded_link(self, embedded_link):
        """Marshal embedded Siren link.

        :param embedded_link: embedded Siren Link.
        :returns: dictionary with embedded link data.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal an embedded link '%s'", embedded_link)

        marshaler = self.create_embedded_link_marshaler(embedded_link)
        try:
            marshaled_link = marshaler.marshal()
        except Exception:
            logger.error("Failed to marshal an embedded link")
            raise

        logger.info("Successfully marshaled an embedded link")
        return marshaled_link

    def marshal_embedded_representation(self, embedded_representation):
        """Marshal Siren embedded representation.

        :param embedded_representation: Siren embedded representation.
        :returns: dictionary with embedded representation data.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal an embedded representation '%s'", embedded_representation)

        marshaler = self.create_embedded_representation_marshaler(embedded_representation)
        try:
            marshaled_representation = marshaler.marshal()
        except Exception:
            logger.error("Failed to marshal an embedded representation")
            raise

        logger.info("Successfully marshaled an embedded representation")
        return marshaled_representation

    def marshal_entity(self, entity):
        """Marshal Siren entity.

        :param entity: Siren entity.
        :returns: dictionary with entity data.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal an entity '%s'", entity)

        marshaler = self.create_entity_marshaler(entity)
        try:
            marshaled_entity = marshaler.marshal()
        except Exception:
            logger.error("Failed to marshal an entity")
            raise

        logger.info("Successfully marshaled an entity")
        return marshaled_entity



"""Module with base class for all Siren components."""

import lila.core.common as common


class Component:
    """Class for base Siren component."""

    def __init__(self, classes=(), title=None):
        self._classes = common.adjust_classes(classes)

        if title is not None:
            title = str(title)
        self._title = title

    @property
    def classes(self):
        """Classes of the component."""
        return tuple(self._classes)

    @property
    def title(self):
        """Descriptive title for the component."""
        return self._title



"""Module with base parser for Siren objects."""


class Parser:
    """Class to parse serialized Siren objects."""

    def parse_field(self, data):
        """Parse serialized Siren field.

        :param data: serialized field.
        :returns: Siren Field.
        """
        raise NotImplementedError("Parser does not support siren fields")

    def parse_action(self, data):
        """Parse serialized Siren action.

        :param data: serialized action.
        :returns: Siren Action.
        """
        raise NotImplementedError("Parser does not support siren actions")

    def parse_link(self, data):
        """Parse serialized Siren link.

        :param data: serialized link.
        :returns: Siren Link.
        """
        raise NotImplementedError("Parser does not support siren links")

    def parse_embedded_link(self, data):
        """Parse serialized embedded Siren link.

        :param data: serialized embedded link.
        :returns: Siren EmbeddedLink.
        """
        raise NotImplementedError("Parser does not support embedded siren links")

    def parse_entity(self, data):
        """Parse serialized Siren entity.

        :param data: serialized entity.
        :returns: Siren Entity.
        """
        raise NotImplementedError("Parser does not support siren entities")

    def parse_embedded_representation(self, data):
        """Parse serialized Siren embedded representation.

        :param data: serialized embedded representation.
        :returns: Siren EmbeddedRepresentation.
        """
        raise NotImplementedError("Parser does not support siren embedded representations")



import os.path
import setuptools


if __name__ == "__main__":
    dir_path = os.path.dirname(__file__)
    readme_path = os.path.abspath(os.path.join(dir_path, "README.rst"))
    with open(readme_path) as readme_file:
        long_description = readme_file.read()

    setuptools.setup(
        name="lila",
        version="1.0.0",
        author="KillAChicken",
        author_email="KillAChicken@yandex.ru",
        description="Library to work with Siren protocol",
        long_description=long_description,
        url="https://github.com/KillAChicken/lila",
        packages=setuptools.find_packages(include=("lila*", )),
        install_requires=[
            ],
        classifiers=[
            "License :: OSI Approved :: MIT License",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            ],
        )



"""Module with default marshaler for an action."""

import logging

from lila.core.action import Action, Method


class ActionMarshaler:
    """Class to marshal a single action."""

    def __init__(self, action, marshaler):
        self._action = action
        self._marshaler = marshaler

    def marshal(self):
        """Marshal the action.

        :returns: dictionary with action data.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to marshal an action")

        action_data = {
            "name": self.marshal_name(),
            "class": self.marshal_classes(),
            "method": self.marshal_method(),
            "href": self.marshal_target(),
            "title": self.marshal_title(),
            "type": self.marshal_media_type(),
            "fields": self.marshal_fields(),
            }

        logger.info("Successfully marshaled an action")
        return action_data

    def marshal_name(self):
        """Marshal action's name.

        :returns: string name of the action.
        :raises: :class:ValueError.
        """
        action = self._action
        try:
            name = action.name
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get action's name")
            raise ValueError("Failed to get action's name") from error

        return str(name)

    def marshal_classes(self):
        """Marshal action's classes.

        :returns: list with string names of action's classes.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        action = self._action
        try:
            classes = list(str(class_) for class_ in action.classes)
        except AttributeError as error:
            logger.error("Failed to get action's classes")
            raise ValueError("Failed to get action's classes") from error
        except TypeError as error:
            logger.error("Failed to iterate over action's classes")
            raise ValueError("Failed to iterate over action's classes") from error

        return classes

    def marshal_method(self):
        """Marshal action's method.

        :returns: string value of action's method.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        action = self._action
        try:
            method = str(action.method)
        except AttributeError as error:
            logger.error("Failed to get action's method")
            raise ValueError("Failed to get action's method") from error

        try:
            method = Method(method)
        except ValueError as error:
            logger.error("Action's method is not supported")
            raise ValueError("Action's method is not supported") from error

        return method.value

    def marshal_target(self):
        """Marshal action's target.

        :returns: string target of the action.
        :raises: :class:ValueError.
        """
        action = self._action
        try:
            target = action.target
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get action's target")
            raise ValueError("Failed to get action's target") from error

        return str(target)

    def marshal_title(self):
        """Marshal action's title.

        :returns: string title of the action or None.
        :raises: :class:ValueError.
        """
        action = self._action
        try:
            title = action.title
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get action's title")
            raise ValueError("Failed to get action's title") from error

        if title is not None:
            title = str(title)

        return title

    def marshal_media_type(self):
        """Marshal action's media type.

        :returns: string value of action's media type or None.
        :raises: :class:ValueError.
        """
        action = self._action
        try:
            media_type = action.media_type
        except AttributeError as error:
            logging.getLogger(__name__).error("Failed to get action's media type")
            raise ValueError("Failed to get action's media type") from error

        if media_type is not None:
            media_type = str(media_type)

        return media_type

    def marshal_fields(self):
        """Marshal action's fields.

        :returns: list with marshaled data of action's fields.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        action = self._action
        try:
            action_fields = list(action.fields)
        except AttributeError as error:
            logger.error("Failed to get action's fields")
            raise ValueError("Failed to get action's fields") from error
        except TypeError as error:
            logger.error("Failed to iterate over action's fields")
            raise ValueError("Failed to iterate over action's fields") from error

        marshal_field = self._marshaler.marshal_field

        marshaled_fields = []
        for field in action_fields:
            try:
                field_data = marshal_field(field)
            except Exception as error:
                logger.error("Failed to marshal action's fields")
                raise ValueError("Failed to marshal action's fields") from error

            marshaled_fields.append(field_data)

        return marshaled_fields


class ActionParser:
    """Class to marshal a single action."""

    def __init__(self, data, parser):
        self._data = data
        self._parser = parser

    def parse(self):
        """Parse an action from the data.

        :returns: :class:`Action <lila.core.action.Action>`.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse an action")

        action_name = self.parse_name()
        action_classes = self.parse_classes()
        action_method = self.parse_method()
        action_target = self.parse_target()
        action_title = self.parse_title()
        action_media_type = self.parse_media_type()
        action_fields = self.parse_fields()

        try:
            action = Action(
                name=action_name,
                classes=action_classes,
                method=action_method,
                target=action_target,
                title=action_title,
                media_type=action_media_type,
                fields=action_fields,
                )
        except Exception as error:
            logger.error("Failed to create an action with provided data")
            raise ValueError("Failed to create an action with provided data") from error
        else:
            logger.info("Successfully parsed an action")

        return action

    def parse_name(self):
        """Parse action's name.

        :returns: string name of the action.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            action_name = self._data["name"]
        except TypeError as error:
            logger.error("Failed to get name from action data")
            raise ValueError("Failed to get name from action data") from error
        except KeyError as error:
            logger.error("Action data do not have required 'name' key")
            raise ValueError("Action data do not have required 'name' key") from error

        return str(action_name)

    def parse_classes(self):
        """Parse action's classes.

        :returns: list with string names of action's classes.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            action_classes = self._data["class"]
        except TypeError as error:
            logger.error("Failed to get classes from action data")
            raise ValueError("Failed to get classes from action data") from error
        except KeyError:
            action_classes = ()

        try:
            action_classes = tuple(str(class_) for class_ in action_classes)
        except TypeError as error:
            logger.error("Failed to iterate over classes from action data")
            raise ValueError("Failed to iterate over classes from action data") from error

        return action_classes

    def parse_method(self):
        """Parse action's method.

        :returns: :class:`Method <lila.core.action.Method>`.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            action_method = self._data["method"]
        except TypeError as error:
            logger.error("Failed to get method from action data")
            raise ValueError("Failed to get method from action data") from error
        except KeyError:
            action_method = Method.GET.value

        try:
            action_method = Method(action_method)
        except ValueError as error:
            logger.error("Action data contain not supported method")
            raise ValueError("Action data contain not supported method") from error

        return action_method

    def parse_target(self):
        """Parse action's target.

        :returns: string target of the action.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            action_target = self._data["href"]
        except TypeError as error:
            logger.error("Failed to get target from action data")
            raise ValueError("Failed to get target from action data") from error
        except KeyError as error:
            logger.error("Action data do not have required 'href' key")
            raise ValueError("Action data do not have required 'href' key") from error

        return str(action_target)

    def parse_title(self):
        """Parse action's title.

        :returns: string title of the action or None.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            action_title = self._data["title"]
        except TypeError as error:
            logger.error("Failed to get title from action data")
            raise ValueError("Failed to get title from action data") from error
        except KeyError:
            action_title = None

        if action_title is not None:
            action_title = str(action_title)

        return action_title

    def parse_media_type(self):
        """Parse action's media type.

        :returns: string value of action's media type or None.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            action_media_type = self._data["type"]
        except TypeError as error:
            logger.error("Failed to get media type from action data")
            raise ValueError("Failed to get media type from action data") from error
        except KeyError:
            action_media_type = None

        if action_media_type is not None:
            action_media_type = str(action_media_type)
        elif self.parse_fields():
            action_media_type = "application/x-www-form-urlencoded"

        return action_media_type

    def parse_fields(self):
        """Parse action's fields.

        :returns: list with parsed action's fields.
        :raises: :class:ValueError.
        """
        logger = logging.getLogger(__name__)

        try:
            action_fields_data = self._data["fields"]
        except TypeError as error:
            logger.error("Failed to get fields data from action data")
            raise ValueError("Failed to get fields data from action data") from error
        except KeyError:
            action_fields_data = ()

        try:
            action_fields_data = list(action_fields_data)
        except TypeError as error:
            logger.error("Failed to iterate over fields data from action data")
            raise ValueError("Failed to iterate over fields data from action data") from error

        parse_field = self._parser.parse_field

        action_fields = []
        for data in action_fields_data:
            try:
                field = parse_field(data)
            except Exception as error:
                logger.error("Failed to parse action's fields")
                raise ValueError("Failed to parse action's fields") from error

            action_fields.append(field)

        return tuple(action_fields)



"""Module with JSON parser for Siren objects."""

import logging

from lila.serialization.parser import Parser
from lila.serialization.json.field import FieldParser
from lila.serialization.json.action import ActionParser
from lila.serialization.json.link import LinkParser, EmbeddedLinkParser
from lila.serialization.json.entity import EntityParser, EmbeddedRepresentationParser


class JSONParser(Parser):
    """Class to parse Siren objects from JSON."""

    create_field_parser = FieldParser
    create_link_parser = LinkParser
    create_embedded_link_parser = EmbeddedLinkParser

    def create_action_parser(self, data):
        """Factory method to create a parser for an action.

        :param data: action data to parse.
        :returns: :class:`ActionParser <lila.serialization.json.action.ActionParser>`.
        """
        return ActionParser(data=data, parser=self)

    def create_entity_parser(self, data):
        """Factory method to create a parser for an entity.

        :param data: entity data to parse.
        :returns: :class:`EntityParser <lila.serialization.json.entity.EntityParser>`.
        """
        return EntityParser(data=data, parser=self)

    def create_embedded_representation_parser(self, data):
        """Factory method to create a parser for an embedded representation.

        :param data: data of embedded representation to parse.
        :returns: :class:`EmbeddedRepresentationParser
            <lila.serialization.json.entity.EmbeddedRepresentationParser>`.
        """
        return EmbeddedRepresentationParser(data=data, parser=self)

    def parse_field(self, data):
        """Parse serialized Siren field.

        :param data: serialized field.
        :returns: parsed field.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse a field from data '%s'", data)

        parser = self.create_field_parser(data)
        try:
            parsed_field = parser.parse()
        except Exception:
            logger.error("Failed to parse a field")
            raise

        logger.info("Successfully parsed a field")
        return parsed_field

    def parse_action(self, data):
        """Parse serialized Siren action.

        :param data: serialized action.
        :returns: parsed action.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse an action from data '%s'", data)

        parser = self.create_action_parser(data)
        try:
            parsed_action = parser.parse()
        except Exception:
            logger.error("Failed to parse an action")
            raise

        logger.info("Successfully parsed an action")
        return parsed_action

    def parse_link(self, data):
        """Parse serialized Siren link.

        :param data: serialized link.
        :returns: parsed link.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse a link from data '%s'", data)

        parser = self.create_link_parser(data)
        try:
            parsed_link = parser.parse()
        except Exception:
            logger.error("Failed to parse a link")
            raise

        logger.info("Successfully parsed a link")
        return parsed_link

    def parse_embedded_link(self, data):
        """Parse serialized Siren embedded link.

        :param data: serialized embedded link.
        :returns: parsed embedded link.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse an embedded link from data '%s'", data)

        parser = self.create_embedded_link_parser(data)
        try:
            parsed_embedded_link = parser.parse()
        except Exception:
            logger.error("Failed to parse an embedded link")
            raise

        logger.info("Successfully parsed an embedded link")
        return parsed_embedded_link

    def parse_embedded_representation(self, data):
        """Parse serialized Siren embedded representation.

        :param data: serialized embedded representation.
        :returns: parsed embedded representation.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse an embedded representation from data '%s'", data)

        parser = self.create_embedded_representation_parser(data)
        try:
            parsed_representation = parser.parse()
        except Exception:
            logger.error("Failed to parse an embedded representation")
            raise

        logger.info("Successfully parsed an embedded representation")
        return parsed_representation

    def parse_entity(self, data):
        """Parse serialized Siren entity.

        :param data: serialized entity.
        :returns: parsed entity.
        """
        logger = logging.getLogger(__name__)
        logger.debug("Try to parse an entity from data '%s'", data)

        parser = self.create_entity_parser(data)
        try:
            parsed_entity = parser.parse()
        except Exception:
            logger.error("Failed to parse an entity")
            raise

        logger.info("Successfully parsed an entity")
        return parsed_entity



"""Module with base marshaler for Siren objects."""


class Marshaler:
    """Class to marshal serialized Siren objects."""

    def marshal_field(self, field):
        """Marshal Siren field.

        :param field: Siren Field.
        :returns: serialized field.
        """
        raise NotImplementedError("Marshaler does not support siren fields")

    def marshal_action(self, action):
        """Marshal Siren action.

        :param action: Siren Action.
        :returns: serialized action.
        """
        raise NotImplementedError("Marshaler does not support siren actions")

    def marshal_link(self, link):
        """Marshal Siren link.

        :param link: Siren Link.
        :returns: dictionary with link data.
        """
        raise NotImplementedError("Marshaler does not support siren links")

    def marshal_embedded_link(self, embedded_link):
        """Marshal embedded Siren link.

        :param embedded_link: embedded Siren Link.
        :returns: dictionary with embedded link data.
        """
        raise NotImplementedError("Marshaler does not support embedded siren links")

    def marshal_embedded_representation(self, embedded_representation):
        """Marshal Siren embedded representation.

        :param embedded_representation: Siren embedded representation.
        :returns: dictionary with embedded representation data.
        """
        raise NotImplementedError("Marshaler does not support siren embedded representation")

    def marshal_entity(self, entity):
        """Marshal Siren entity.

        :param entity: Siren entity.
        :returns: dictionary with entity data.
        """
        raise NotImplementedError("Marshaler does not support siren entity")
